#
# treeview.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   you can distribute/modify this program under the terms of
#   the GNU General Public License version 2 or later.
#

require 'delta/view'


module Delta

  class TreeView < View

    def initialize( ident, mbox )
      @identifier = ident
      @mbox = mbox
      @tree = Delta::MailThreadTreeGtk.new
      init_window
      @multiselect = false

      mbox.attach self
    end

    attr_reader :identifier

    CACHE_VERSION = 1

    def init2__parse
      begin_update
      @tree.add_port_set @mbox
      end_update
      show_tree
    end

    def init2__cached( f )
raise
    end

    def init2__cache_failed
    end

    def finalize__cache( f )
    end

    def inspect
      "#<#{type} #{@mbox.inspect}>"
    end

    def mailbox
      @mbox
    end

    def size
      @tree.size
    end

    def empty?
      @tree.empty?
    end

    def add( port )
      return unless view_initialized?
      @tree.add_port port
    end

    def del( port )
      return unless view_initialized?
      @tree.remove_port port
    end

    def clear
      return unless view_initialized?
      n = @tree.size
      @tree.clear
      n
    end

    def move_selection_to( dest )
      # must be view_initialized
      @mbox.move_ports_to dest, selection()
    end

    def selection
      node = @tree.selected or return []

      if @multiselect then
        set = []
        node.each_node do |n|
          set.push n.port
        end
        set
      else
        [ node.port ]
      end
    end

    def multiselect
      @multiselect
    end

    alias multiselect? multiselect

    def multiselect=( bool )
      @multiselect = bool
    end

    def toggle_multiselect
      self.multiselect = !multiselect()
    end


    def window
      @viewport
    end

    def adjustment
      @viewport.get_vadjustment
    end

    def begin_update
    end

    def end_update
      @tree.start_update
    end


    def focus_next
      @tree.focus_next
    end

    def focus_prev
      @tree.focus_prev
    end

    private

    def init_window
      @viewport = Gtk::Viewport.new( nil, @adjustment ) 
      @viewport.set_usize 640, 300
    end
      
    def show_tree
      @viewport.add @tree.window
      @tree.window.show_all
    end
  
  end


  #
  # Toplevel Tree
  #
  #    (structure)
  #
  #    -- toplevel tree (grand root)
  #         -- root
  #             -- node
  #             -- node
  #                 -- node
  #         -- root
  #             -- node
  #

  class MailThreadTree

    def initialize( rootc, nodec )
      @node_class = nodec
      @root_class = rootc

      @roots     = []
      @id2node   = {}
      @port2node = {}
      @pending   = {}
      @orphan    = {}
    end

    def size
      @id2node.size
    end

    def empty?
      @id2node.empty?
    end

    def clear
      @roots.clear
      @id2node.clear
      @pending.clear
      @orphan.clear
    end

    def remove_port( port, whole = false )
puts 'TreeView: removing port: ' + port.inspect
      node = @port2node[ port ] or return
      del node
      if node.root? then
        is_root = true
        delete_root node.parent, node
      else
        is_root = false
        parent = node.tearoff
      end

      if whole then
        # remove also children from tree
        node.children.each do |c|
          c.each_node do |n|
            del n
          end
        end
      else
        # re-connect
        if is_root then
          # becomes root
          node.children.each do |c|
            add_node_as_root c
          end
        else
          node.children.each do |c|
            parent.add_child c
          end
        end
      end
    end

    def del( node )
      @orphan.delete node
      @port2node.delete node.port
      @id2node.delete node.message_id
    end
    private :del

    def add_port_set( portset )
      adding {
        portset.each do |port|
          process_mail port
        end
      }
    end

    def add_port( port )
      adding {
        process_mail port
      }
    end

    def start_update   # called from TreeView
    end

    private

    def adding
      @tmp_roots = []
      yield
      a = @tmp_roots.compact
      a.each do |node|
        add_node_as_root node
      end
      @tmp_roots.clear
      @orphan.clear
    end

    def add_node_as_root( node )
      node.set_is_root true
      r = @root_class.new( self, TMail::Mail.new(node.port) )
      r.add_child node
      r.set_parent self
      @roots.push r
      r
    end

    def delete_root( root, real )
      @roots.delete root
      real.set_is_root false
      real.set_parent nil
    end

    def process_mail( port )
      node = @node_class.new( self, TMail::Mail.new(port) )
      @id2node[ node.message_id ] = node
      @port2node[ port ] = node

      if orphan? node then
        (@pending[ node.parent_id ] ||= []).push node
        @orphan[ node.message_id ] = @tmp_roots.size
        @tmp_roots.push node
        return
      end

      if parent = parent_of(node) then
        parent.add_child node
      else
        @tmp_roots.push node
      end
      check_pending node
    end

    def orphan?( node )
      node.parent_id and not @id2node.key? node.parent_id
    end

    def parent_of( node )
      @id2node[ node.parent_id ]
    end

    def check_pending( parent )
      a = @pending.delete( parent.message_id ) or return
      a.each do |child|
        i = @orphan.delete( child.message_id )
        @tmp_roots[i] = nil if i
        child.tearoff if child.parent
        parent.add_child child
      end
    end

    #
    # node search
    #

    def next_node_of( node )
      # child
      n = node.children[0]
      n and return n

      while node do
        # brother
        if node.root? then
          return @roots[ @roots.index(node.parent) + 1 ]
        end
        n = node.parent.next_child_of(node)
        n and return n

        node = node.parent
      end

      nil
    end

  end


  class MailThreadTreeGtk < MailThreadTree

    def initialize
      super MailThreadRootGtk, MailThreadNodeGtk
      @tree = Gtk::Tree.new
      @tree.selection_mode = Gtk::SELECTION_BROWSE
      @selected = nil
    end

    attr_reader :tree
    alias window tree

    attr_accessor :selected

    def clear
      @tree.clear_items 0, -1
      @tree = Gtk::Tree.new
      @tree.selection_mode = Gtk::SELECTION_BROWSE
      super
    end

    def start_update   # called from TreeView
      @roots.each do |r|
        r.gui_construct
      end
    end

    def delete_root( root, real )
      # root.treeitem = nil before gui_construct().
      root.gui_clear
      super
    end
    private :delete_root

    def focus_next
      n = next_node_of(@selected)
      n.select
    end

    def focus_prev
    end
  
  end


  #
  # Node
  #

  class MailThreadNode

    def initialize( groot, mail )
      @grand_root = groot
      @port = mail.port
      @message_id = mail.message_id(nil)
      @parent_id , = mail.in_reply_to(nil)

      @parent = nil
      @children = []
      @is_root = false

      @n_children = 0
    end

    attr_reader :port

    attr_reader :children
    attr_reader :parent

    attr_reader :n_children

    attr_reader :message_id
    attr_reader :parent_id

    def set_is_root( flag )
      @is_root = flag
    end

    def root?
      @is_root
    end

    def root
      if @is_root then
        @parent
      else
        @parent ? @parent.root : nil
      end
    end

    def inspect
      "#<#{self.class} #{__id__}>"
    end

    alias to_s inspect

    def each_node( &block )
      yield self
      @children.each do |c|
        c.each_node( &block )
      end
    end

    def add_child( node )
      @children.push node
      @n_children += (1 + node.n_children)
      @parent.recompute_n_children if @parent
      node.set_parent self
    end

    def set_parent( node )
      @parent and raise ArgumentError, 'parent is set twice'
      @parent = node
    end

    def recompute_n_children
      @n_children = 0
      @children.each do |n|
        @n_children += (1 + n.n_children)
      end
      @parent.recompute_n_children if @parent
    end

    def remove_child( node )
      @children.delete node or return false
      @n_children -= (1 + node.n_children)
      # node.parent = nil
      true
    end

    def tearoff
      parent = @parent
      parent.remove_child self
      parent
    end

    def next_child_of( c )
      @children[ @children.index(c) + 1 ]
    end

  end


  class MailThreadNodeGtk < MailThreadNode
  
    def initialize( groot, mail )
      super
      @label = get_label(mail)
      @tree = nil
      @treeitem = nil
    end

    def tree
      unless @tree
        @tree = Gtk::Tree.new
        @tree.selection_mode = Gtk::SELECTION_BROWSE
        @treeitem.subtree = @tree
      end
      @tree
    end

    attr_reader :treeitem

    def gui_construct
      return if @treeitem

      @treeitem = Gtk::TreeItem.new(@label)
      sighandler @treeitem
      @parent.tree.append @treeitem
      @treeitem.show
      @children.each do |c|
        c.gui_construct
      end
      @treeitem.expand
    end

    def remove_child( node )
      super or return false
      node.gui_clear
      if @children.empty? then
        @tree = nil
      end
      true
    end

    def gui_clear
      @children.each do |c|
        c.gui_clear
      end
      @tree = nil
      @parent.tree.remove_item @treeitem
      @treeitem = nil
    end

    def select
      @parent.tree.select_child @treeitem
    end
    
    def deselect
      @treeitem.deselect
    end

    private

    def get_label( mail )
      mail.strftime(' %m/%d ', '') + mail.friendly_from('')
    end

    def sighandler( i )
      i.signal_handler('select') {|node|
          @grand_root.selected = self
          Delta.mailer.view @port
      }
    end

  end


  #
  # Root (label node)
  #

  class MailThreadRootGtk < MailThreadNodeGtk

    def gui_construct
      return if @treeitem

      @label_base = @label
      @label = "(#{@n_children}) " + @label_base
      super
      @treeitem.collapse
    end

    def recompute_n_children
      super
p @treeitem.label if @treeitem
    end

    private

    def get_label( mail )
      mail.subject.gsub(/\s+/, ' ').sub(/\[[^:]+:(\d+)\]/, '\1')
    end

    def sighandler( i )
      i.signal_handler( 'select' ) {|item|
          item.expand
          @grand_root.selected = self
      }
      # selecting children causes to deselect root
      #i.signal_handler( 'deselect' ) {|node|
      #    node.collapse
      #}
    end
  
  end

end
